<?php

namespace Drupal\openstack\Service;

use Drupal\aws_cloud\Entity\Ec2\ImageInterface;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceException;
use Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface;
use Drupal\aws_cloud\Traits\AwsCloudEntityCheckTrait;
use Drupal\cloud\Entity\CloudConfigInterface;
use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\cloud\Traits\CloudContentEntityTrait;
use Drupal\Core\Entity\EntityDeleteFormTrait;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\Messenger;
use Drupal\openstack\Service\Rest\OpenStackService as OpenStackRestService;

/**
 * Interacts with OpenStack using OpenStackService API.
 */
class OpenStackOperationsService implements OpenStackOperationsServiceInterface {

  use AwsCloudEntityCheckTrait;
  use CloudContentEntityTrait;
  use EntityDeleteFormTrait;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  private $cloudConfigPluginManager;

  /**
   * The Cloud service.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  private $cloudService;

  /**
   * The AWS Cloud EC2 service.
   *
   * @var \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface
   */
  private $ec2Service;

  /**
   * The entity being used by this form.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface|\Drupal\Core\Entity\RevisionLogInterface
   */
  private $entity;

  /**
   * The Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private $entityTypeManager;

  /**
   * The OpenStack Service Factory.
   *
   * @var \Drupal\openstack\Service\OpenStackServiceFactoryInterface
   */
  private $openStackServiceFactory;

  /**
   * Ec2OperationsService constructor.
   *
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The Cloud service.
   * @param \Drupal\aws_cloud\Service\Ec2\Ec2ServiceInterface $ec2_service
   *   The AWS Cloud EC2 service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The Entity Type Manager.
   * @param \Drupal\Core\Messenger\Messenger $messenger
   *   The messenger service.
   * @param \Drupal\openstack\Service\OpenStackServiceFactoryInterface $openstack_service_factory
   *   Object for interfacing with OpenStack Service.
   */
  public function __construct(
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    CloudServiceInterface $cloud_service,
    Ec2ServiceInterface $ec2_service,
    EntityTypeManagerInterface $entity_type_manager,
    Messenger $messenger,
    OpenStackServiceFactoryInterface $openstack_service_factory
  ) {
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->cloudService = $cloud_service;
    $this->ec2Service = $ec2_service;
    $this->entityTypeManager = $entity_type_manager;
    $this->messenger = $messenger;
    $this->openStackServiceFactory = $openstack_service_factory;
  }

  /**
   * Gets the form entity.
   */
  public function getEntity() {
    return $this->entity;
  }

  /**
   * Trim white spaces in the values of textfields.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  private function trimTextfields(array $form, FormStateInterface $form_state): void {
    $field_names = [];
    foreach ($form ?: [] as $name => $item) {
      if (!is_array($item)) {
        continue;
      }

      if (isset($item['#type'])
        && $item['#type'] === 'textfield'
      ) {
        $field_names[] = $name;
      }

      if (isset($item['#type']) && $item['#type'] === 'details') {
        foreach ($item ?: [] as $sub_item_name => $sub_item) {
          if (is_array($sub_item)
            && isset($sub_item['#type'])
            && $sub_item['#type'] === 'textfield'
          ) {
            $field_names[] = $sub_item_name;
          }
        }
      }
    }

    foreach ($field_names ?: [] as $field_name) {
      $value = $form_state->getValue($field_name);
      if ($value === NULL) {
        continue;
      }

      $value = trim($value);
      $form_state->setValue($field_name, $value);
      $this->entity->set($field_name, $value);
    }
  }

  /**
   * Check if current cloud config is remote or not.
   *
   * @return bool
   *   True if the current cloud config is remote.
   */
  private function isCloudConfigRemote(): bool {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface
     */
    $entity = $this->entity;

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    return $cloud_config->isRemote();
  }

  /**
   * Set the tags for the entity.
   *
   * @param string $resource_id
   *   The resource ID.  For example, instance_id, volume_id.
   * @param array $tag_map
   *   The map of tags.
   * @param string $entity_type
   *   The entity type.
   */
  private function setTagsInAws($resource_id, array $tag_map, $entity_type = ''): void {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface
     */
    $entity = $this->entity;

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $tags = [];
    foreach ($tag_map ?: [] as $key => $value) {
      $tags[] = [
        'Key' => $key,
        'Value' => $value,
      ];
    }

    if (!method_exists($this->ec2Service, 'createTags')) {
      return;
    }

    // Create Tags with different parameters for AWS and OpenStack.
    if (preg_match('[^aws_cloud]', $entity->getEntityTypeId()) === 1) {
      $this->ec2Service->createTags([
        'Resources' => [$resource_id],
        'Tags' => $tags,
      ]);
    }
    else {
      $this->ec2Service->createTags([
        'Resources' => [$resource_id],
        'Tags' => $tags,
        'EntityType' => $entity_type,
      ]);
    }
  }

  /**
   * Update name and created_by tags.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param string $aws_cloud_resource_id
   *   The AWS Cloud resource ID.
   */
  private function updateNameAndCreatedByTags(EntityInterface $entity, $aws_cloud_resource_id): void {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface|\Drupal\aws_cloud\Entity\Ec2\Image
     */
    $original_entity = $entity;

    $cloud_context = $original_entity->getCloudContext();
    if ($this->isCloudConfigRemote()) {
      // Get original cloud context.
      if (($pos = strrpos($cloud_context, '_')) !== FALSE) {
        $cloud_context = substr($cloud_context, 0, $pos);
      }
    }

    $bundle = str_contains($original_entity->bundle(), 'aws_cloud')
      ? 'aws_cloud' : 'openstack';
    $uid_key_name = $this->cloudService->getTagCreatedByUid($bundle, $cloud_context);
    $this->setTagsInAws($aws_cloud_resource_id, [
      $original_entity->getEntityTypeId() . '_' . $uid_key_name => !empty($original_entity->getOwner())
      ? $original_entity->getOwner()->id() : 0,
      'Name' => $original_entity->getName(),
    ],
    $original_entity->getEntityTypeId());
  }

  /**
   * Create an Image.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return bool
   *   TRUE when the process succeeds.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function createImage(array $form, FormStateInterface $form_state): bool {
    $this->trimTextfields($form, $form_state);

    /** @var \Drupal\aws_cloud\Entity\Ec2\Image $entity */
    $entity = $this->entity;

    $result = $this->ec2Service->createImage([
      'InstanceId'  => $entity->getInstanceId(),
      'Name'        => $entity->getName(),
      'Description' => $entity->getDescription(),
      'Visibility'  => $entity->getVisibility(),
    ]);

    if (!empty($result['SendToWorker'])) {
      $this->processOperationStatus($entity, 'created remotely');
      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      return TRUE;
    }

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $account_id = $this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value;

    if (!empty($entity) && !empty($result['ImageId'])
      && ($entity->setName($form_state->getValue('name')))
      && ($entity->set('ami_name', $form_state->getValue('name')))
      && ($entity->setImageId($result['ImageId']))
      && ($entity->set('account_id', $account_id))
      && ($entity->save())) {

      $this->updateNameAndCreatedByTags($entity, $entity->getImageId());

      $this->processOperationStatus($entity, 'created');

      $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
      $this->dispatchSaveEvent($entity);

      return TRUE;
    }
    else {
      $this->processOperationErrorStatus($entity, 'created');

      return FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;

      // Switch OpenStack service based on $entity->getCloudContext().
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());

      $this->createImage($form, $form_state);

      if (!empty($result['SendToWorker'])) {
        return TRUE;
      }

      if ($this->ec2Service instanceof OpenStackRestService) {
        $this->updateNameAndCreatedByTags($entity, $entity->getImageId());
      }

      return TRUE;
    }
    catch (Ec2ServiceException $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Copy values from #type=item elements to its original element type.
   *
   * @param array $form
   *   The form array.
   */
  private function copyFormItemValues(array $form): void {
    /**
     * @var \Drupal\Core\Entity\ContentEntityInterface
     */
    $original_entity = $this->entityTypeManager
      ->getStorage($this->entity->getEntityTypeId())
      ->load($this->entity->id());

    $item_field_names = [];
    foreach ($form ?: [] as $name => $item) {
      if (!is_array($item)) {
        continue;
      }

      if (isset($item['#type'])
        && $item['#type'] === 'item'
        && (!isset($item['#not_field']) || $item['#not_field'] === FALSE)
      ) {
        $item_field_names[] = $name;
      }

      if (isset($item['#type']) && $item['#type'] === 'details') {
        foreach ($item ?: [] as $sub_item_name => $sub_item) {
          if (is_array($sub_item)
            && isset($sub_item['#type'])
            && $sub_item['#type'] === 'item'
            && (!isset($sub_item['#not_field']) || $sub_item['#not_field'] === FALSE)
          ) {
            $item_field_names[] = $sub_item_name;
          }
        }
      }
    }

    foreach ($item_field_names ?: [] as $item_field_name) {
      // Support multi-valued item fields.
      $values = !empty($original_entity) ? $original_entity->get($item_field_name)->getValue() : [];
      if (!empty($values) && count($values) > 1) {
        $item_field_values = array_map(static function ($value) {
          return $value['value'];
        }, $values);
        $this->entity->set($item_field_name, $item_field_values);
      }

      $item_field_value = !empty($original_entity) && !empty($original_entity->get($item_field_name))
        ? $original_entity->get($item_field_name)->value
        : [];
      if (!empty($item_field_value)) {
        $this->entity->set($item_field_name, $item_field_value);
      }
    }
  }

  /**
   * Save a cloud content.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveCloudContent(array $form, FormStateInterface $form_state) {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface
     */
    $entity = $this->entity;

    if (!($entity instanceof CloudConfigInterface)) {
      $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
      $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();
    }

    if (!$entity->isNew() && !($entity instanceof CloudConfigInterface) && $cloud_config->isRemote()) {
      $status = SAVED_UPDATED;
      $passive_operation = 'updated remotely';
      $this->processOperationStatus($entity, $passive_operation);
      return $status;
    }

    $status = $entity->save();

    // Add a updated status message and the log.
    switch ($status) {
      case SAVED_NEW:
        $passive_operation = 'created';
        break;

      case SAVED_DELETED:
        $passive_operation = 'deleted';
        break;

      default:
      case SAVED_UPDATED:
        $passive_operation = 'updated';
    }

    if (!($entity instanceof CloudConfigInterface) && $cloud_config->isRemote()) {
      $passive_operation .= ' remotely';
    }

    $this->processOperationStatus($entity, $passive_operation);

    // Clear block and menu cache.
    $this->clearCacheValues($entity->getCacheTags());

    return $status;
  }

  /**
   * Save an AWS Cloud content.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveAwsCloudContent(array $form, FormStateInterface $form_state): void {
    // Copy values of the form elements whose type are item to entity.
    // If not, the properties corresponding to the form elements
    // will be saved as NULL.
    $this->copyFormItemValues($form);

    $this->trimTextfields($form, $form_state);

    $this->saveCloudContent($form, $form_state);
  }

  /**
   * Save an Image.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  private function saveImage(array $form, FormStateInterface $form_state): void {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface|\Drupal\aws_cloud\Entity\Ec2\Image
     */
    $entity = $this->entity;

    /**
     * The original entity data.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface|\Drupal\aws_cloud\Entity\Ec2\Image
     */
    $original_entity = $this->entityTypeManager
      ->getStorage($entity->getEntityTypeId())
      ->load($entity->id());

    $this->saveAwsCloudContent($form, $form_state);

    $this->updateNameAndCreatedByTags($entity, $entity->getImageId());

    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $cloud_config = $this->cloudConfigPluginManager->loadConfigEntity();

    // Call 'modifyImageAttribute'
    // only when OwnerId is same as Account ID of Cloud Config.
    // It's to avoid AuthFailure error from AWS API.
    if ($cloud_config->get('field_account_id')->value !== $entity->getAccountId()) {
      return;
    }

    $form_values = $form_state->getValues();
    $visibility = $form_values['visibility'] ?? 0;

    $permission = [
      [
        'Group' => 'all',
        'UserId' => $entity->getAccountId(),
      ],
    ];

    // 1 => Public, 0 => Private.
    $launch_permission =
      $visibility
        ? ['Add' => $permission]
        : ['Remove' => $permission];

    // Update image.
    $this->ec2Service->setCloudContext($entity->getCloudContext());

    if (!method_exists($this->ec2Service, 'modifyImageAttribute')) {
      return;
    }

    // Update image description.
    $this->ec2Service->modifyImageAttribute([
      'ImageId' => $entity->getImageId(),
      'Description' => ['Value' => $entity->getDescription()],
    ]);

    // These fields (Description , LaunchPermission , or ProductCode) cannot
    // be specified at the same time.
    // See also:
    // https://docs.aws.amazon.com/cli/latest/reference/ec2/modify-image-attribute.html
    // Update image launch permission.
    $this->ec2Service->modifyImageAttribute([
      'ImageId' => $entity->getImageId(),
      'LaunchPermission' => $launch_permission,
    ]);

    $new_account_ids = [];
    foreach ($entity->getLaunchPermissionAccountIds() as $account_id) {
      if (empty($account_id->value)) {
        continue;
      }
      $new_account_ids[] = $account_id->value;
    }

    $old_account_ids = [];
    foreach ($original_entity->getLaunchPermissionAccountIds() as $account_id) {
      if (empty($account_id->value)) {
        continue;
      }
      $old_account_ids[] = $account_id->value;
    }

    $account_ids_to_remove = array_diff($old_account_ids, $new_account_ids);
    $account_ids_to_add = array_diff($new_account_ids, $old_account_ids);

    if (!empty($account_ids_to_remove)) {
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => [
          'Remove' => array_map(function ($item) {
            return ['UserId' => $item];
          }, $account_ids_to_remove),
        ],
      ]);
    }

    if (!empty($account_ids_to_add)) {
      $this->ec2Service->modifyImageAttribute([
        'ImageId' => $entity->getImageId(),
        'LaunchPermission' => [
          'Add' => array_map(function ($item) {
            return ['UserId' => $item];
          }, $account_ids_to_add),
        ],
      ]);
    }

    $this->clearCacheValues($entity->getCacheTags());
    $this->dispatchSaveEvent($entity);
  }

  /**
   * {@inheritdoc}
   */
  public function editOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      $this->saveImage($form, $form_state);

      /**
       * The entity being used by this form.
       *
       * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface|\Drupal\aws_cloud\Entity\Ec2\Image
       */
      $original_entity = $entity;
      $visibility = !empty($original_entity->getVisibility()) ? 'public' : 'private';

      // Update Image of OpenStack REST API.
      if ($this->ec2Service instanceof OpenStackRestService) {
        // Make an API call to update image.
        $result = $this->ec2Service->updateImage([
          'ImageId' => $original_entity->getImageId(),
          'Name' => $original_entity->getName(),
          'Visibility' => $visibility,
        ]);

        $old_tags = $result['Tags'] ?? [];

        foreach ($old_tags ?: [] as $tag) {
          // Delete old tags.
          $this->ec2Service->deleteTags(
            [
              'ImageId'    => $original_entity->getImageId(),
              'Tag'        => $tag,
              'EntityType' => $original_entity->getEntityTypeId(),
            ]);
        }

        $this->updateNameAndCreatedByTags($original_entity, $original_entity->getImageId());
      }

      return TRUE;
    }
    catch (Ec2ServiceException $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

  /**
   * Delete an Image.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  private function deleteImage(array &$form, FormStateInterface $form_state): bool {
    /**
     * The entity being used by this form.
     *
     * @var \Drupal\aws_cloud\Entity\Ec2\ImageInterface|\Drupal\aws_cloud\Entity\Ec2\Image
     */
    $entity = $this->entity;
    if ($entity->getEntityTypeId() === 'aws_cloud_image'
        && !$this->updateEntitySubmitForm($form_state, $entity)) {
      return FALSE;
    }

    $this->ec2Service->setCloudContext($entity->getCloudContext());
    $this->cloudConfigPluginManager->setCloudContext($entity->getCloudContext());
    $account_id = $this->cloudConfigPluginManager->loadConfigEntity()->get('field_account_id')->value;

    // If the image is not owned by the aws user,
    // the calling for deregisterImage will be skipped for AWS Cloud
    // and it will be called for OpenStack.
    if (($entity->getEntityTypeId() === 'aws_cloud_image'
    && ($entity->getAccountId() !== $account_id))
    || ($result = $this->ec2Service->deregisterImage([
      'ImageId' => $entity->getImageId(),
    ])) !== NULL) {

      if (!empty($result['SendToWorker'])) {
        $this->messenger->addStatus($this->t('The @type @label has been deleted remotely.', [
          '@type'  => $entity->getEntityType()->getSingularLabel(),
          '@label' => $entity->getName(),
        ]));

        $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
        return TRUE;
      }

      $entity->delete();

      // Do not change the following message since we cannot use
      // $entity->label(), which represents 'ami_name'.
      $this->messenger->addStatus($this->t('The @type @label has been deleted.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));
      $this->logDeletionMessage();
      $this->clearCacheValues($entity->getCacheTags());
      $this->dispatchSubmitEvent($entity);
    }
    else {

      // Do not change the following message since we cannot use
      // $entity->label(), which represents 'ami_name'.
      $this->messenger->addError($this->t('The @type @label could not be deleted.', [
        '@type'  => $entity->getEntityType()->getSingularLabel(),
        '@label' => $entity->getName(),
      ]));

      $this->logger($entity->getEntityType()->getProvider())->error($this->t('@type: @label could not be deleted.', [
        '@type' => $entity->getEntityType()->getLabel(),
        '@label' => $entity->label(),
      ]), [
        'link' => $entity->toLink($this->t('View'))->toString(),
      ]);
    }

    $form_state->setRedirect("view.{$entity->getEntityTypeId()}.list", ['cloud_context' => $entity->getCloudContext()]);
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteOpenStackImage(ImageInterface $entity, array &$form, FormStateInterface $form_state): bool {
    try {
      $this->entity = $entity;
      $this->ec2Service = $this->openStackServiceFactory->get($entity->getCloudContext());
      return $this->deleteImage($form, $form_state);
    }
    catch (Ec2ServiceException $e) {
      $this->handleException($e);
      return FALSE;
    }
  }

}
